#include "globals.h"
#include "DepthMap.h"
#include "plyio.h"
#include "meshAux.h"
#include "softRender.h"
#include "xfio.h"

int
main(int argc, char **argv)
{
   float xmin, xmax, ymin, ymax;
   int xdim, ydim;

   // Read in a transformation matrix and print it's contents
   Matrix4f cameraPose;
   readXform(argv[1], cameraPose);
   cameraPose.print();

   // Initialize and print out the marching cubes table
   initEdgeTable();
   FILE *fp = fopen(argv[2], "w");
   printEdgeTable(fp);
   fclose(fp);

   // Read in a range image
   RangeGrid *rangeGrid = new RangeGrid;
   rangeGrid->readPly(argv[3]);

   // Convert the range image to a mesh (a.k.a., a range surface).
   // The triangles that get created should have a normal with a
   // z-component greater than 0.1
   Mesh *mesh = meshFromGrid(rangeGrid, 1, USE_VIEW_DOT_NORMAL, 0.1);

   // Write range surface to file
   mesh->writePly(argv[4]);


   // Prepare to rasterize the range surface into a depth map.  This
   // example is trivial and assumes an orthographic projection along
   // the z-axis with image pixel spacing of unity.  In general, you
   // would need to specify the desired pixel spacing commensurate
   // with the resolution of the model and build the projection and
   // z-buffer accordingly.

   // Find bounding box
   mesh->computeBBox();

   // Copy the min/max of bounding box into more sensibly named variables
   xmin = mesh->bbox.fll.x;
   xmax = mesh->bbox.nur.x;
   ymin = mesh->bbox.fll.y;
   ymax = mesh->bbox.nur.y;

   // Compute the appropriate depth map dimensions
   xdim = int(xmax-xmin)+1;
   ydim = int(ymax-ymin)+1;

   // Create a projector object.  The current projector object just
   // stores a matrix and its inverse and applies these as the
   // projection and un-projection process.  This is fine for
   // orthographic projection, but you would need to change this if
   // you wanted to work accurately with data acquired via a
   // perspective projection.

   Projector *proj = new Projector;
   proj->mat.makeIdentity();
   proj->mat.setTranslate(-xmin, -ymin, 0);
   proj->matInv.setValue(proj->mat.inverse());
   
   // Apply the projection operation to the mesh before rasterizing.
   mesh->applyXform(proj->mat);

   // Set up a depth map
   DepthMap *depthMap = new DepthMap(xdim,ydim);
   depthMap->proj = proj;

   // Rasterize the mesh into the depth map
   softRender(depthMap, mesh);

   // Convert the depth map into a range grid and write it out to a file.
   RangeGrid *resampGrid = depthMap->buildRangeGrid();
   resampGrid->writePly(argv[5]);
}

