#ifndef _MC_H_
#define _MC_H_

#include "Mesh.h"

/******************************************************************

Here's how the voxels and edges in a cube are enumerated:


                    v7______e6_____v6
                    /|             /|
                   / |            / |
                e10  |          e11 |
                 /  e7          /   e5
                /    |         /    |
              v3_______e2_____v2    |
               |     |        |     |
               |     |        |     |
               |    v4_____e4_|____v5
               |    /         |    /
              e3   /         e1   /
               |  e8          |  e9
               | /            | /
               |/             |/
              v0______e0______v1
         

Since each voxel can take on either positive or negative value, and
there are 8 voxels to a cube, then there are 2^8=256 cases to
consider.  The edge table will have 256 entries, one for each case.
Indexing into the edge table is determined by setting or clearing a
bit depending on whether the voxel is positive or negative.  Here's a
sample voxel sign enumeration:

  Voxel    Sign
    0       -      
    1       +
    2       +
    3       +
    4       +
    5       +
    6       -
    7       +

If each voxel corresponds to a bit, and + sets a bit, while - clears
it, then we have the bit code 10111110 = 190.  This configuration of
signs should cause a triangle to be generated just inside of voxel 0
and another just inside of voxel 6.  The triangle triangle new voxel 0
should use zero crossings at e0,e3,e8 (following the right hand rule).
The triangle triangle new voxel 6 should use zero crossings at
e5,e6,e11 (following the right hand rule).  If you look at Lorensen's
table, you'll see that he starts his counting of cases at zero, but
his counting of edges at one.  So, we need to look up case 190 and
then subtract one from each of his edges.  Entry 190 says:

  { 1, 4, 9, 0, 6, 7, 12, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}

which means that triangles are formed from zero crossings on edge
triples 1,4,9 and 6,7,12 in Lorensen's nomenclature.  After
subtracting 1 from each, we get 0,3,8 and 5,6,12, just as we expect.

For a given case, the edge table will explicity state which edges have
a sign crossing, the number of triangles that pass through the cube,
and how the triangles are formed from the set of enumerated edge
crossings.

*******************************************************************/

typedef struct {
   float value;
   Vec3f gradient;
   float confidence;
   float intensity;
   uchar matDiff[3];
} Voxel;

typedef Voxel Cube[8];

typedef struct {
   int edgeBool[12];    // 0 or 1 for each edge
   int numTris;         // Number of triangles for a given case
   Triangle tris[5];    // Triangles as edge triples, at most 5 triangles
} EdgeTableEntry;


void initEdgeTable();
void printEdgeTable(FILE *fp);

#endif

