#include <stdio.h>
#include <ply.h>
#include <stdlib.h>
//changed by lizhang
#ifdef LINUX
#include <strings.h>
#endif
#ifdef WIN32
#include <string.h>
#endif
//end of change

#include "meshAux.h"
#include "globals.h"
#include "defines.h"

static void addTriangleUseNormal(Mesh *mesh, int vin1, int vin2, int vin3, 
				 float minDot, Vec3f &viewDir);
static void addTriangleUseLength(Mesh *mesh, int vin1, int vin2, int vin3, 
				 float maxLength);
void reallocTris(Vertex *v);
void reallocVerts(Vertex *v);
int vertexEdgeTest(Mesh *mesh, Vertex *vert);
void addNeighbors(Vertex *v1, Vertex *v2);



/******************************************************************************

meshFromGrid():
  Create a triangle mesh from range grid data.

Entry:
  rangeGrid -- the range grid data
  subSamp -- can subsample the range grid by a factor of subSamp if desired
  criterion -- criterion deciding when to create triangles from neighboring 
               range values
  thresh -- threshold for decision based on criterion

Exit:
  returns pointer to newly-created mesh

******************************************************************************/

Mesh *
meshFromGrid(RangeGrid *rangeGrid, int subSamp, int criterion, float thresh)
{
   int i,j;
   Vec3f v1, v2;
   int ii,jj;
   int in1,in2,in3,in4,vin1,vin2,vin3,vin4;
   Vertex *vt1,*vt2,*vt3,*vt4;
   int count;
   int max_lt,max_lg;
   Mesh *mesh;
   int nlt,nlg,max_verts,max_tris;
   int *vert_index;
   
   mesh = new Mesh;
   
   /* allocate space for new triangles and vertices */
   
   mesh->numVerts = 0;
   nlt = rangeGrid->nlt;
   nlg = rangeGrid->nlg;
   max_lt = (nlt - 1) / subSamp + 1;
   max_lg = (nlg - 1) / subSamp + 1;
   max_verts = max_lt * max_lg;
   
   /* create a list saying whether a vertex is going to be used */
   
   vert_index = new int[rangeGrid->numSamples];
   for (i = 0; i < rangeGrid->numSamples; i++)
      vert_index[i] = -1;
   
   /* see which vertices will be used in a triangle */
   
   count = 0;
   for (i = 0; i <= nlt - subSamp; i += subSamp)
      for (j = 0; j <= nlg - subSamp; j += subSamp) {
	 in1 = rangeGrid->indices[i + j * nlt];
	 if (in1 >= 0) {
	    vert_index[in1] = 1;
	    count++;
	 }
      }
   
   mesh->numVerts = count;
   mesh->verts = new Vertex[mesh->numVerts];
   
   mesh->numTris = 0;
   max_tris = mesh->numVerts * 2;
   mesh->tris = new Triangle[max_tris];
   
   
   /* create the vertices */
   
   count = 0;
   for (i = 0; i < rangeGrid->numSamples; i++) {
      if (vert_index[i] == -1)
	 continue;
      
      vert_index[i] = count;
      mesh->verts[count].coord = rangeGrid->coords[i];
      
      mesh->verts[count].maxVerts = 8;
      mesh->verts[count].verts = new Vertex*[8];
      mesh->verts[count].edgeLengths = new float[8];
      mesh->verts[count].numVerts = 0;
      
      mesh->verts[count].maxTris = 8;
      mesh->verts[count].tris = new Triangle*[8];
      mesh->verts[count].numTris = 0;
      
      if (rangeGrid->hasConfidence)
	 mesh->verts[count].confidence = rangeGrid->confidence[i];
      else 
	 mesh->verts[count].confidence = 1;
      
      count++;
   }
   mesh->numVerts = count;
   
   /* create the triangles */
   
   for (i = 0; i < nlt - subSamp; i += subSamp)
      for (j = 0; j < nlg - subSamp; j += subSamp) {
	 
	 ii = (i + subSamp) % nlt;
	 jj = (j + subSamp) % nlg;
	 
	 in1 = rangeGrid->indices[ i +  j * nlt];
	 in2 = rangeGrid->indices[ii +  j * nlt];
	 in3 = rangeGrid->indices[ii + jj * nlt];
	 in4 = rangeGrid->indices[ i + jj * nlt];
	 
	 count = (in1 >= 0) + (in2 >= 0) + (in3 >= 0) + (in4 >=0);
	 if (in1 >= 0) {
	    vin1 = vert_index[in1];
	    vt1 = &mesh->verts[vin1];
	    vt1->stepsToEdge = 0;
	 }
	 if (in2 >= 0) {
	    vin2 = vert_index[in2];
	    vt2 = &mesh->verts[vin2];
	    vt2->stepsToEdge = 0;
	 }
	 if (in3 >= 0) {
	    vin3 = vert_index[in3];
	    vt3 = &mesh->verts[vin3];
	    vt3->stepsToEdge = 0;
	 }
	 if (in4 >= 0) {
	    vin4 = vert_index[in4];
	    vt4 = &mesh->verts[vin4];
	    vt4->stepsToEdge = 0;
	 }
	 
	 switch (criterion) {
	   case USE_EDGE_LENGTH :
	      if (count == 4) {    /* all 4 vertices okay, so make 2 tris */
		 float len1,len2;
		 
		 /* compute lengths of cross-edges */
		 v1 = vt1->coord - vt3->coord;
		 v2 = vt2->coord - vt4->coord;
		 len1 = v1.length();
		 len2 = v2.length();
		 
		 /* make triangles that minimize the cross-length */
		 
		 if (len1 < len2) {
		    addTriangleUseLength(mesh, vin1, vin2, vin3, thresh);
		    addTriangleUseLength(mesh, vin1, vin3, vin4, thresh);
		 } else {
		    addTriangleUseLength(mesh, vin2, vin3, vin4, thresh);
		    addTriangleUseLength(mesh, vin2, vin4, vin1, thresh);
		 }
	      }
	      else if (count == 3) {   /* only 3 vertices okay, so make 1 tri */
		 if (in1 == -1) {
		    addTriangleUseLength(mesh, vin2, vin3, vin4, thresh);
		 }
		 else if (in2 == -1) {
		    addTriangleUseLength(mesh, vin1, vin3, vin4, thresh);
		 }
		 else if (in3 == -1) {
		    addTriangleUseLength(mesh, vin1, vin2, vin4, thresh);
		 }
		 else {	/* in4 == -1 */
		    addTriangleUseLength(mesh, vin1, vin2, vin3, thresh);
		 }
	      }
	      break;

	      
	   case USE_VIEW_DOT_NORMAL :
	      if (count == 4) {    /* all 4 vertices okay, so make 2 tris */
		 float len1,len2;
		 
		 /* compute lengths of cross-edges */
		 v1 = vt1->coord - vt3->coord;
		 v2 = vt2->coord - vt4->coord;
		 len1 = v1.length();
		 len2 = v2.length();
		 
		 /* make triangles that minimize the cross-length */
		 
		 if (len1 < len2) {
		    addTriangleUseNormal(mesh, vin1, vin2, vin3, 
					 thresh, rangeGrid->viewDir);
		    addTriangleUseNormal(mesh, vin1, vin3, vin4, 
					 thresh, rangeGrid->viewDir);
		 } else {
		    addTriangleUseNormal(mesh, vin2, vin3, vin4, 
					 thresh, rangeGrid->viewDir);
		    addTriangleUseNormal(mesh, vin2, vin4, vin1, 
					 thresh, rangeGrid->viewDir);
		 }
	      }
	      else if (count == 3) {   /* only 3 vertices okay, so make 1 tri */
		 if (in1 == -1) {
		    addTriangleUseNormal(mesh, vin2, vin3, vin4, 
					 thresh, rangeGrid->viewDir);
		 }
		 else if (in2 == -1) {
		    addTriangleUseNormal(mesh, vin1, vin3, vin4, 
					 thresh, rangeGrid->viewDir);
		 }
		 else if (in3 == -1) {
		    addTriangleUseNormal(mesh, vin1, vin2, vin4, 
					 thresh, rangeGrid->viewDir);
		 }
		 else {	/* in4 == -1 */
		    addTriangleUseNormal(mesh, vin1, vin2, vin3, 
					 thresh, rangeGrid->viewDir);
		 }
	      }
	 }
      }
   
   /* free up the vertex index list */
   delete [] vert_index;
   
   mesh->initNormals();  
   
   findMeshEdges(mesh);
   
   return (mesh);
}


void
addConnectivity(Mesh *mesh)
{
   int i;
   for (i = 0; i < mesh->numVerts; i++) {
      mesh->verts[i].stepsToEdge = 0;    
      mesh->verts[i].holeFill = FALSE;    
      mesh->verts[i].numVerts = 0;    
      mesh->verts[i].numTris = 0;    
      mesh->verts[i].maxVerts = 8;
      mesh->verts[i].verts = new Vertex*[8];
      mesh->verts[i].edgeLengths = new float[8];
      mesh->verts[i].maxTris = 8;
      mesh->verts[i].tris = new Triangle*[8];
   }
   
   for (i = 0; i < mesh->numTris; i++) {
      Triangle *tri = &mesh->tris[i];
      
      Vertex *vert1 = &mesh->verts[tri->vindex1];
      Vertex *vert2 = &mesh->verts[tri->vindex2];
      Vertex *vert3 = &mesh->verts[tri->vindex3];
      
      if (vert1->numTris == vert1->maxTris)
	 reallocTris(vert1);
      
      if (vert2->numTris == vert2->maxTris)
	 reallocTris(vert2);
      
      if (vert3->numTris == vert3->maxTris)
	 reallocTris(vert3);
      
      vert1->tris[vert1->numTris++] = tri;
      vert2->tris[vert2->numTris++] = tri;
      vert3->tris[vert3->numTris++] = tri;
      
      addNeighbors(vert1,vert2);
      addNeighbors(vert1,vert3);
      addNeighbors(vert2,vert3);
      
   }
}

static void
addTriangleUseNormal(Mesh *mesh, int vin1, int vin2, int vin3, 
		     float minDot, Vec3f &viewDir)
{
    Vec3f c1, c2, c3;
    Vec3f v1, v2, v3;
    Triangle *tri;
    Vertex *vert1, *vert2, *vert3;
    int tooGrazing;
    float dot;

    c1.setValue(mesh->verts[vin1].coord);
    c2.setValue(mesh->verts[vin2].coord);
    c3.setValue(mesh->verts[vin3].coord);

    v1 = c1 - c2;
    v2 = c1 - c3;
    v3 = v2.cross(v1);
    v3.normalize();
    dot = v3.dot(viewDir);

    tooGrazing = dot < minDot;
    if (tooGrazing)
	return;

    // No check on limits here???
    tri = &mesh->tris[mesh->numTris];

    tri->vindex1 = vin1;
    tri->vindex2 = vin2;
    tri->vindex3 = vin3;

    vert1 = &mesh->verts[vin1];
    vert2 = &mesh->verts[vin2];
    vert3 = &mesh->verts[vin3];

    vert1->tris[vert1->numTris++] = tri;
    vert2->tris[vert2->numTris++] = tri;
    vert3->tris[vert3->numTris++] = tri;

    addNeighbors(vert1,vert2);
    addNeighbors(vert1,vert3);
    addNeighbors(vert2,vert3);


    mesh->numTris++;
}

static void
addTriangleUseLength(Mesh *mesh, int vin1, int vin2, int vin3, 
		     float maxLength)
{
    Vec3f c1, c2, c3;
    Vec3f v;
    Triangle *tri;
    Vertex *vert1, *vert2, *vert3;
    int tooLong;

    c1.setValue(mesh->verts[vin1].coord);
    c2.setValue(mesh->verts[vin2].coord);
    c3.setValue(mesh->verts[vin3].coord);

    v = c1 - c2;
    tooLong = v.length() > maxLength;
    if (tooLong)
	return;

    v = c1 - c3;
    tooLong = tooLong || v.length() > maxLength;
    if (tooLong)
	return;

    v = c2 - c3;
    tooLong = tooLong || v.length() > maxLength;
    if (tooLong)
	return;

    // No check on limits here???
    tri = &mesh->tris[mesh->numTris];

    tri->vindex1 = vin1;
    tri->vindex2 = vin2;
    tri->vindex3 = vin3;

    vert1 = &mesh->verts[vin1];
    vert2 = &mesh->verts[vin2];
    vert3 = &mesh->verts[vin3];

    vert1->tris[vert1->numTris++] = tri;
    vert2->tris[vert2->numTris++] = tri;
    vert3->tris[vert3->numTris++] = tri;

    addNeighbors(vert1,vert2);
    addNeighbors(vert1,vert3);
    addNeighbors(vert2,vert3);

    if (tooLong) {
	vert1->holeFill = 1;
	vert2->holeFill = 1;
	vert3->holeFill = 1;
	vert1->stepsToEdge = -1;
	vert2->stepsToEdge = -1;
	vert3->stepsToEdge = -1;
    }

    mesh->numTris++;
}


void
addNeighbors(Vertex *v1, Vertex *v2)
{
    float length;
    Vec3f a;
    int found = 0;

    for (int i = 0; i < v1->numVerts; i++) {
	if (v1->verts[i] == v2) {
	    found = 1;
	    break;
	}
    }

    if (!found) {
	length = 0.0005;
	a = v1->coord - v2->coord;
	length = a.length();

	if (v1->numVerts == v1->maxVerts)
	   reallocVerts(v1);
	
	if (v2->numVerts == v2->maxVerts)
	   reallocVerts(v2);
	
	v1->edgeLengths[v1->numVerts] = length;
	v2->edgeLengths[v2->numVerts] = length;
	v1->verts[v1->numVerts++] = v2;
	v2->verts[v2->numVerts++] = v1;
    }
}


/******************************************************************************
Find which vertices of a mesh are on the edge of the mesh.

Entry:
  mesh - mesh to find edges of
******************************************************************************/

void
findMeshEdges(Mesh *mesh)
{
  int i;

  /* examine each vertex of a mesh to see if it's on the edge */

  for (i = 0; i < mesh->numVerts; i++)
    vertexEdgeTest (mesh, &mesh->verts[i]);
}


/******************************************************************************
Mark a vertex to say whether it is on the edge of a mesh or not.

Entry:
  vert: vertex to mark as on edge or not

Exit:
  returns 0 if everything was okay, 1 if the mesh is build funny
******************************************************************************/

int 
vertexEdgeTest(Mesh *mesh, Vertex *vert)
{
  int i;
  Triangle *tri;
  unsigned char on_edge;
  static int been_here = 0;
  int bad_mesh = 0;

  /* initialize list that counts how many times an edge has been marked */
  /* (each edge should be marked twice, once for each triangle, */
  /*  otherwise the vertex is on an edge) */

  for (i = 0; i < vert->numVerts; i++)
    vert->verts[i]->count = 0;

  on_edge = 0;  /* assume we're not on an edge */

  /* go through each triangle of vertex, counting how many times an */
  /* adjacent vertex has been used */

  for (i = 0; i < vert->numTris; i++) {
    tri = vert->tris[i];

    /* count the participation of each vertex of the triangle */
    mesh->verts[tri->vindex1].count++;
    mesh->verts[tri->vindex2].count++;
    mesh->verts[tri->vindex3].count++;


    // This prevents us from operating on them as vertices
    // internal to the mesh
    if (mesh->verts[tri->vindex1].holeFill)
	on_edge = 1;

    if (mesh->verts[tri->vindex2].holeFill)
	on_edge = 1;

    if (mesh->verts[tri->vindex3].holeFill)
	on_edge = 1;
  }

  /* examine the counts of the neighboring vertices to see if they */
  /* prove the vertex to be an edge.  also check for mesh consistancy */

  for (i = 0; i < vert->numVerts; i++) {
    if (vert->verts[i]->count == 1) {
      on_edge = 1;
    }
    else if (vert->verts[i]->count == 2) {
      /* this is okay */
    }
    else {
      /* if we're here, our mesh is built wrong */
      bad_mesh = 1;
      if (!been_here) {
	been_here = 1;
	fprintf (stderr, "vertex_edge_test: %d count on an edge\n",
		 vert->verts[i]->count);
        /*
	fprintf (stderr, "(You'll only get this message once, so beware!)\n");
        */
      }
    }
  }

  vert->on_edge = on_edge;

  /* say if the mesh is improperly built */
  return (bad_mesh);
}



void
reallocVerts(Vertex *v)
{
    int i;
    Vertex **newVerts;
    float *newLengths;

    v->maxVerts *= 2;
    newVerts = new Vertex*[v->maxVerts];
    newLengths = new float[v->maxVerts];
    for (i = 0; i < v->numVerts; i++) {
        newVerts[i] = v->verts[i];
        newLengths[i] = v->edgeLengths[i];
    }
    delete [] v->verts;
    delete [] v->edgeLengths;

    v->verts = newVerts;
    v->edgeLengths = newLengths;
}


void
reallocTris(Vertex *v)
{
    int i;
    Triangle **newTris;

    v->maxTris *= 2;
    newTris = new Triangle*[v->maxTris];
    for (i = 0; i < v->numTris; i++) {
        newTris[i] = v->tris[i];
    }
    delete [] v->tris;

    v->tris = newTris;
}

